import time
from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Optional

import aiosmtplib
import httpx

from nonebot import logger
from .notice_strategy import NoticeStrategy
from .utils import AsHttpReq, sc_send, feishu_gen_sign


class MailStrategy(NoticeStrategy):
    """邮件通知策略"""

    def check_params(self) -> bool:
        return bool(self.config.user and self.config.password and 
                   self.config.server and self.config.port and 
                   self.config.notice_email)

    async def send(self, title: str, content: str) -> Optional[str]:
        try:
            message = MIMEMultipart("alternative")
            message["Subject"] = Header(title, 'utf-8')
            message["From"] = self.config.user
            message["To"] = self.config.notice_email
            message.attach(MIMEText(content))

            use_tls = self.config.port == 465
            async with aiosmtplib.SMTP(
                hostname=self.config.server, 
                port=self.config.port, 
                use_tls=use_tls
            ) as smtp:
                await smtp.login(self.config.user, self.config.password)
                await smtp.send_message(message)

            logger.info("通知邮件发送成功!")
            return None
        except Exception as e:
            err = f"邮件发送失败,错误信息如下{e}"
            logger.error(err)
            return err


class PushPlusStrategy(NoticeStrategy):
    """PushPlus通知策略"""

    def check_params(self) -> bool:
        return bool(self.config.token)

    async def send(self, title: str, content: str) -> Optional[str]:
        try:
            url = "http://www.pushplus.plus/send"
            json_data = {
                "token": self.config.token,
                "title": title,
                "content": content,
                "template": "txt",
                "channel": "wechat"
            }

            r = await AsHttpReq.post(url, json=json_data)
            if r.status_code == 200:
                r_json = r.json()
                if r_json["code"] == 200:
                    logger.info("pushplus通知成功!")
                    return None
                else:
                    err = f"pushplus通知失败,res:{r.text}"
                    logger.error(err)
                    return err
            else:
                err = f"pushplus通知失败,res:{r.text}"
                logger.error(err)
                return err
        except (httpx.ConnectError, httpx.ConnectTimeout):
            err = "pushplus通知失败,bot服务器网络错误"
            logger.error(err)
            return err
        except Exception as e:
            err = f"pushplus通知失败,错误信息如下{e}"
            logger.error(err)
            return err


class ServerStrategy(NoticeStrategy):
    """Server酱Turbo通知策略"""

    def check_params(self) -> bool:
        return bool(self.config.key)

    async def send(self, title: str, content: str) -> Optional[str]:
        try:
            options = {"noip": 1, "channel": 9}
            r = await sc_send(
                sendkey=self.config.key, 
                title=title, 
                desp=content, 
                options=options
            )

            if r.status_code == 200:
                r_json = r.json()
                if r_json["code"] == 0:
                    logger.info("server酱Turbo通知成功!")
                    return None
                else:
                    err = f"server酱Turbo通知失败,res:{r.text}"
                    logger.error(err)
                    return err
            else:
                err = f"server酱Turbo通知失败,res:{r.text}"
                logger.error(err)
                return err
        except (httpx.ConnectError, httpx.ConnectTimeout):
            err = "server酱Turbo通知失败,bot服务器网络错误"
            logger.error(err)
            return err
        except Exception as e:
            err = f"server酱Turbo通知失败,错误信息如下{e}"
            logger.error(err)
            return err


class Server3Strategy(NoticeStrategy):
    """Server酱3通知策略"""

    def check_params(self) -> bool:
        return bool(self.config.key)

    async def send(self, title: str, content: str) -> Optional[str]:
        try:
            r = await sc_send(
                sendkey=self.config.key, 
                title=title, 
                desp=content
            )

            if r.status_code == 200:
                r_json = r.json()
                if r_json["code"] == 0:
                    logger.info("server酱3通知成功!")
                    return None
                else:
                    err = f"server酱3通知失败,res:{r.text}"
                    logger.error(err)
                    return err
            else:
                err = f"server酱3通知失败,res:{r.text}"
                logger.error(err)
                return err
        except (httpx.ConnectError, httpx.ConnectTimeout):
            err = "server酱3通知失败,bot服务器网络错误"
            logger.error(err)
            return err
        except Exception as e:
            err = f"server酱3通知失败,错误信息如下{e}"
            logger.error(err)
            return err


class PushOverStrategy(NoticeStrategy):
    """PushOver通知策略"""

    def check_params(self) -> bool:
        return bool(self.config.token and self.config.user_key)

    async def send(self, title: str, content: str) -> Optional[str]:
        try:
            url = "https://api.pushover.net/1/messages.json"
            json_data = {
                "token": self.config.token,
                "user": self.config.user_key,
                "title": title,
                "message": content
            }

            r = await AsHttpReq.post(url, json=json_data)
            if r.status_code == 200:
                r_json: dict = r.json()
                if r_json["status"] == 1 and (r_json.get("info") is None):
                    logger.info("pushover通知成功!")
                    return None
                else:
                    err = f"pushover通知失败,res:{r.text}"
                    logger.error(err)
                    return err
            else:
                err = f"pushover通知失败,res:{r.text}"
                logger.error(err)
                return err
        except (httpx.ConnectError, httpx.ConnectTimeout):
            err = "pushover通知失败,bot服务器网络错误"
            logger.error(err)
            return err
        except Exception as e:
            err = f"pushover通知失败,错误信息如下{e}"
            logger.error(err)
            return err


class FeiShuStrategy(NoticeStrategy):
    """飞书通知策略"""

    def check_params(self) -> bool:
        return bool(self.config.webhook_url)

    async def send(self, title: str, content: str) -> Optional[str]:
        try:
            webhook_url = self.config.webhook_url
            secret = self.config.is_have_secret()

            body = {
                "msg_type": "text",
                "content": {
                    "text": f"{title}\n{content}"
                }
            }

            if secret:
                timestamp = str(int(time.time()))
                sign = feishu_gen_sign(timestamp, secret)
                body["timestamp"] = timestamp
                body["sign"] = sign

            r = await AsHttpReq.post(webhook_url, json=body)
            if r.status_code == 200:
                r_json = r.json()
                if r_json["code"] == 0:
                    logger.info("飞书通知成功!")
                    return None
                else:
                    err = f"飞书通知失败,res:{r.text}"
                    logger.error(err)
                    return err
            else:
                err = f"飞书通知失败,res:{r.text}"
                logger.error(err)
                return err
        except (httpx.ConnectError, httpx.ConnectTimeout):
            err = "飞书通知失败,bot服务器网络错误"
            logger.error(err)
            return err
        except Exception as e:
            err = f"飞书通知失败,错误信息如下{e}"
            logger.error(err)
            return err



import base64
import hashlib
import hmac
import re

import httpx
from httpx import Response

HTTP_TIME_OUT = 10  # 请求超时，秒

PLATFORM_PUSHPLUS = "pushplus"
PLATFORM_MAIL = "mail"
PLATFORM_SERVER = "server"
PLATFORM_SERVER3 = "server3"
PLATFORM_PUSHOVER = "pushover"
PLATFORM_FEISHU = "feishu"


class AsHttpReq(object):
    """httpx 异步请求封装"""

    @staticmethod
    async def get(url, **kwargs) -> Response:
        proxy = None
        async with httpx.AsyncClient(proxy=proxy) as client:
            response = await client.get(url, timeout=HTTP_TIME_OUT, **kwargs)
            return response

    @staticmethod
    async def post(url, **kwargs) -> Response:
        proxy = None
        async with httpx.AsyncClient(proxy=proxy) as client:
            response = await client.post(url, timeout=HTTP_TIME_OUT, **kwargs)
            return response


async def sc_send(sendkey, title, desp='', options=None):
    """发送消息，兼容Server酱Turbo和Server酱³"""
    if options is None:
        options = {}
    if sendkey.startswith('sctp'):
        # server3
        match = re.match(r'^sctp(\d+)t', sendkey)
        if match:
            url = f'https://{match.group(1)}.push.ft07.com/send/{sendkey}.send'
        else:
            raise ValueError("Invalid sendkey format for 'sctp'.")
    else:
        # server turbo
        url = f'https://sctapi.ftqq.com/{sendkey}.send'
    params = {
        'title': title,
        'desp': desp,
        **options
    }
    headers = {
        'Content-Type': 'application/json;charset=utf-8'
    }
    response = await AsHttpReq.post(url, json=params, headers=headers)
    return response


def feishu_gen_sign(timestamp, secret):
    """飞书生成加密校验"""
    # 拼接timestamp和secret
    string_to_sign = '{}\n{}'.format(timestamp, secret)
    hmac_code = hmac.new(string_to_sign.encode("utf-8"), digestmod=hashlib.sha256).digest()
    # 对结果进行base64处理
    sign = base64.b64encode(hmac_code).decode('utf-8')
    return sign



from nonebot import logger, Bot
from .config import plugin_config
from .dataClass import MailConfig, BotParams, PushPlusConfig, ServerConfig, PushOverConfig, Server3Config, FeiShuConfig
from .utils import *
from .strategies import (
    MailStrategy,
    PushPlusStrategy,
    ServerStrategy,
    Server3Strategy,
    PushOverStrategy,
    FeiShuStrategy
)

# 配置实例
mail_config: MailConfig = MailConfig(
    user=plugin_config.disconnect_notice_smtp_user,
    password=plugin_config.disconnect_notice_smtp_password,
    server=plugin_config.disconnect_notice_smtp_server,
    port=plugin_config.disconnect_notice_smtp_port,
    notice_email=plugin_config.disconnect_notice_notice_email,
)

pushplus_config: PushPlusConfig = PushPlusConfig(
    token=plugin_config.disconnect_notice_pushplus_token,
)

server_config: ServerConfig = ServerConfig(
    key=plugin_config.disconnect_notice_server_key,
)

server3_config: Server3Config = Server3Config(
    key=plugin_config.disconnect_notice_server3_key,
)

pushover_config: PushOverConfig = PushOverConfig(
    user_key=plugin_config.disconnect_notice_pushover_user_key,
    token=plugin_config.disconnect_notice_pushover_token
)

feishu_config: FeiShuConfig = FeiShuConfig(
    webhook_url=plugin_config.disconnect_notice_feishu_webhook_url,
    secret=plugin_config.disconnect_notice_feishu_secret,
)


class NoticeContext:
    """通知上下文,管理所有通知策略"""
    
    def __init__(self):
        self._strategies = {
            PLATFORM_MAIL: MailStrategy(mail_config),
            PLATFORM_PUSHPLUS: PushPlusStrategy(pushplus_config),
            PLATFORM_SERVER: ServerStrategy(server_config),
            PLATFORM_SERVER3: Server3Strategy(server3_config),
            PLATFORM_PUSHOVER: PushOverStrategy(pushover_config),
            PLATFORM_FEISHU: FeiShuStrategy(feishu_config),
        }
    
    def get_strategy(self, platform: str):
        """获取指定平台的通知策略"""
        return self._strategies.get(platform)
    
    def get_all_strategies(self):
        """获取所有通知策略"""
        return self._strategies


# 创建全局通知上下文实例
notice_context = NoticeContext()


async def send_notice(bot_params: BotParams, test=False):
    """整合发送通知消息"""
    mode_list = plugin_config.disconnect_notice_mode_list
    
    # 构建消息主体
    if not test:
        title = f"[nonebot2]你的bot掉线了"
        if not bot_params.tag:
            content = f"你的 {bot_params.adapter_name} 适配器的bot账号: {bot_params.bot_id} 掉线了，可能是被风控了，赶快去看看吧"
        else:
            content = f"你的 {bot_params.adapter_name} 适配器的bot账号: {bot_params.bot_id} 掉线了，原因为: {bot_params.tag} 。赶快去看看吧"
    else:
        title = f"[nonebot2]掉线通知测试"
        content = f"这是一个掉线通知测试信息，你的bot并没有掉线"
    
    # 使用策略模式发送通知
    last_error = None
    for platform in mode_list:
        strategy = notice_context.get_strategy(platform)
        if strategy:
            err = await strategy.send_with_check(title, content)
            if err:
                logger.error(err)
                last_error = err
    
    return last_error



from abc import ABC, abstractmethod
from typing import Optional
from nonebot import logger


class NoticeStrategy(ABC):
    """通知策略的抽象基类"""

    def __init__(self, config):
        self.config = config

    @abstractmethod
    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        pass

    @abstractmethod
    async def send(self, title: str, content: str) -> Optional[str]:
        """发送通知
        Returns:
            Optional[str]: 发送失败时返回错误信息,成功返回None
        """
        pass

    def get_platform_name(self) -> str:
        """获取平台名称"""
        return self.__class__.__name__.replace("Strategy", "").lower()

    async def send_with_check(self, title: str, content: str) -> Optional[str]:
        """发送通知(带参数检查)"""
        if not self.check_params():
            err = f"{self.get_platform_name()} 平台通知缺少配置参数,无法进行通知,\n" \
                  "请参考https://github.com/Cypas/nonebot_plugin_disconnect_notice#%EF%B8%8F-%E9%85%8D%E7%BD%AE\n" \
                  "填写该推送方式配置项"
            logger.error(err)
            return err

        return await self.send(title, content)



import json
from datetime import datetime as dt, timedelta

from nonebot.internal.matcher import Matcher
from nonebot.plugin import PluginMetadata
from nonebot import Bot, require, get_bots, on_notice
from nonebot import logger

# onebot11 协议
from nonebot.adapters.onebot.v11 import Bot as V11Bot
from nonebot.adapters.onebot.v11 import Event as V11Event
from nonebot import on_command
from nonebot.permission import SUPERUSER

from .config import driver, plugin_config, global_config, Config
from .dataClass import BotParams
from .notice import send_notice, notice_context
from .utils import PLATFORM_SERVER, PLATFORM_MAIL, PLATFORM_PUSHPLUS, PLATFORM_PUSHOVER, PLATFORM_SERVER3, PLATFORM_FEISHU

require("nonebot_plugin_apscheduler")
from nonebot_plugin_apscheduler import scheduler

__plugin_meta__ = PluginMetadata(
    name="bot断连通知",
    description="bot断连时的通知插件，当前支持邮件通知，pushplus微信公众号通知，server酱公众号通知，pushover多平台设备通知",
    usage="""
    超级用户指令:
        /掉线测试:主动触发掉线通知测试
    """,
    type="application",
    # 发布必填，当前有效类型有：`library`（为其他插件编写提供功能），`application`（向机器人用户提供功能）。
    homepage="https://github.com/Cypas/nonebot_plugin_disconnect_notice",
    # 发布必填。
    config=Config,
    supported_adapters=None,
)

notice_test = on_command("断连通知测试",aliases={"掉线测试","掉线通知测试"}, priority=8, block=True, permission=SUPERUSER)


@notice_test.handle()
async def _(matcher: Matcher):
    """主动触发掉线通知测试"""
    msg = "已发送测试通知\n"
    mode_list = plugin_config.disconnect_notice_mode_list
    # 使用策略模式检查各平台的配置
    for platform in mode_list:
        strategy = notice_context.get_strategy(platform)
        if strategy and strategy.check_params():
            if platform == PLATFORM_MAIL:
                msg += "\n当前通知渠道包括mail，若邮件未收到时请检查邮件垃圾箱"
            elif platform == PLATFORM_PUSHOVER:
                msg += "\n当前通知渠道包括pushover，若未收到通知时，请检查是否有任一平台(安卓/ios/ipad/pc/mac)的app保持了后台运行且打开了系统通知"
    
    bot_params = BotParams(
        adapter_name="114514",
        bot_id="114514"
    )
    res = await send_notice(bot_params, test=True)
    if res is not None:
        msg = f"测试通知发送失败，请重新检查配置项参数正确性，错误信息为: {res}"
    await matcher.finish(msg)

# NapCatQQ与Lagrange.Core下线情况单独处理
offline = on_notice(priority=1, block=False)
@offline.handle()
async def _(bot: V11Bot,event: V11Event):
    if event.get_event_name() == "notice.bot_offline":
        des = json.loads(event.get_event_description().replace("'", '"'))
        if des:
            # 开发者模式下不生效
            platform = "NapCatQQ或Lagrange"
            bot_id = bot.self_id
            if not plugin_config.disconnect_notice_dev_mode:
                job_id = f"disconnect_notice_{platform}_{bot_id}"
                if scheduler.get_job(job_id):
                    scheduler.remove_job(job_id)
                # 计算运行时间
                run_time = dt.now() + timedelta(seconds=plugin_config.disconnect_notice_max_grace_time)

                bot_params = BotParams(
                    adapter_name=platform,
                    bot_id=bot_id,
                    tag = des['tag'],
                )
                scheduler.add_job(
                    id=job_id, func=cron_send_notice, args=[bot_params],
                    misfire_grace_time=60, coalesce=True, max_instances=1, trigger='date', run_date=run_time
                )
            else:
                logger.info(f"当前掉线通知已开启dev模式，不会进行任何实际通知")


@driver.on_bot_disconnect
async def disconnect(bot: Bot):
    """bot断连触发器"""
    # 开发者模式下不生效
    platform = bot.adapter.get_name()
    bot_id = bot.self_id
    if not plugin_config.disconnect_notice_dev_mode:
        job_id = f"disconnect_notice_{platform}_{bot_id}"
        if scheduler.get_job(job_id):
            scheduler.remove_job(job_id)
        # 计算运行时间
        run_time = dt.now() + timedelta(seconds=plugin_config.disconnect_notice_max_grace_time)

        bot_params = BotParams(
            adapter_name=platform,
            bot_id=bot_id
        )
        scheduler.add_job(
            id=job_id, func=cron_send_notice, args=[bot_params],
            misfire_grace_time=60, coalesce=True, max_instances=1, trigger='date', run_date=run_time
        )
    else:
        logger.info(f"当前掉线通知已开启dev模式，不会进行任何实际通知")

async def cron_send_notice(bot_params: BotParams):
    """定时任务：发送通知"""
    platform =bot_params.adapter_name
    bot_id = bot_params.bot_id
    # 重新检查bot是否接入
    bots = get_bots()
    bot = bots.get(bot_id)
    if (not bot) or (platform == "NapCatQQ或Lagrange"):
        logger.warning(f"平台:{platform} bot_id:{bot_id} 已掉线，即将进行通知")
        await send_notice(bot_params)
    else:
        logger.info(f"平台:{platform} bot_id:{bot_id} 已重连，无需进行通知")


@driver.on_bot_connect
async def connect(bot: Bot):
    """bot接入时检测配置完整性"""
    if isinstance(bot, V11Bot):
        mode_list = plugin_config.disconnect_notice_mode_list
        # 使用策略模式检查所有平台的配置
        for platform in mode_list:
            strategy = notice_context.get_strategy(platform)
            if strategy and not strategy.check_params():
                # 缺少参数,私聊通知主人
                super_user: str = list(global_config.superusers)[0]
                if super_user:
                    msg = f"【插件nonebot-plugin-disconnect-notice】\n缺少{platform}配置项，请按照" \
                          "https://github.com/Cypas/nonebot_plugin_disconnect_notice#%EF%B8%8F-%E9%85%8D%E7%BD%AE" \
                          "\n添加配置项后，再重新载入插件"
                    await bot.send_private_msg(user_id=int(super_user), message=msg)
                    break



from abc import ABC, abstractmethod


class BaseConfig(ABC):
    """配置基类,定义所有配置类的共同接口"""
    
    @abstractmethod
    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        pass


class MailConfig(BaseConfig):
    def __init__(self,
                 user: str,
                 password: str,
                 server: str,
                 port: int,
                 notice_email: str):
        self.user = user
        self.password = password
        self.server = server
        self.port = port
        self.notice_email = notice_email

    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        return bool(self.user and self.password and self.server and self.port and self.notice_email)


# https://www.pushplus.plus/doc/
class PushPlusConfig(BaseConfig):
    def __init__(self, token: str):
        self.token = token

    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        return bool(self.token)


# https://open.feishu.cn/document/client-docs/bot-v3/add-custom-bot?f_link_type=f_linkinlinenote&flow_extra=eyJpbmxpbmVfZGlzcGxheV9wb3NpdGlvbiI6MCwiZG9jX3Bvc2l0aW9uIjowLCJkb2NfaWQiOiJiNGU5MmI3OGM2NDBkYzk5LTIzNmNkY2VmNmQ1NGFmOTYifQ%3D%3D&lang=zh-CN
class FeiShuConfig(BaseConfig):
    def __init__(self, webhook_url: str, secret=""):
        self.webhook_url = webhook_url
        self.secret = secret

    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        return bool(self.webhook_url)

    def is_have_secret(self):
        """是否使用了请求鉴权"""
        return self.secret


# https://sct.ftqq.com/
class ServerConfig(BaseConfig):
    def __init__(self, key: str):
        self.key = key

    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        return bool(self.key)


# https://sct.ftqq.com/
class Server3Config(BaseConfig):
    def __init__(self, key: str):
        self.key = key

    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        return bool(self.key)


# https://pushover.net/api#messages
class PushOverConfig(BaseConfig):
    def __init__(self, user_key: str, token: str):
        self.user_key = user_key
        self.token = token

    def check_params(self) -> bool:
        """检查参数是否填写完整"""
        return bool(self.token and self.user_key)


class BotParams:
    def __init__(self,
                 adapter_name: str,
                 bot_id: str,
                 tag: str = None,
                 ):
        self.adapter_name = adapter_name
        self.bot_id = bot_id
        self.tag = tag



from typing import Union

from nonebot import get_driver, get_plugin_config
from pydantic import BaseModel


# 其他地方出现的类似 from .. import config，均是从 __init__.py 导入的 Config 实例
class Config(BaseModel):
    # 推送方式list 枚举值:pushplus mail server等
    disconnect_notice_mode_list: list[str] = ["server"]

    # 邮件推送
    disconnect_notice_smtp_user: Union[str, int] = ""
    disconnect_notice_smtp_password: Union[str, int] = ""
    disconnect_notice_smtp_server: Union[str, int] = ""
    disconnect_notice_smtp_port: int = 465
    disconnect_notice_notice_email: Union[str, int] = ""

    # pushplus推送 https://www.pushplus.plus/
    disconnect_notice_pushplus_token: str = ""

    # server酱Turbo https://sct.ftqq.com/r/1483
    disconnect_notice_server_key: str = ""

    # server酱3 https://sc3.ft07.com/
    disconnect_notice_server3_key: str = ""

    # pushover
    disconnect_notice_pushover_user_key: str = ""
    disconnect_notice_pushover_token: str = ""

    # feishu
    disconnect_notice_feishu_webhook_url: str = ""
    disconnect_notice_feishu_secret: str = ""

    # 其他设定
    disconnect_notice_dev_mode: bool = False # dev开发模式，此模式下仅输出日志，不实际进行通知
    disconnect_notice_max_grace_time: int = 10 # 断连时最大宽限时间:秒 在此期间重连便不会进行通知


driver = get_driver()
# 本地测试时由于不启动 driver，需要将下面两行注释并取消再下一行的注释
global_config = driver.config
plugin_config = get_plugin_config(Config)

# plugin_config = Config()
